@@ -3,6 +3,9 @@ from collections import Counter
 import nltk
 from tqdm import tqdm
 import torch
+import numpy as np
+import matplotlib.pyplot as plt
+import pandas as pd
 
 
 def intersection_over_union(boxes_preds, boxes_labels, box_format="midpoint"):
@@ -56,7 +59,9 @@ def intersection_over_union(boxes_preds, boxes_labels, box_format="midpoint"):
     return intersection / (box1_area + box2_area - intersection + 1e-6)
 
 
-def mean_average_precision(pred_boxes, true_boxes, iou_threshold=0.5, box_format="midpoint"):
+def mean_average_precision(
+    pred_boxes, true_boxes, iou_threshold=0.5, box_format="midpoint", interpolate_pr=True
+):
     """
     Calculates mean average precision
     Parameters:
@@ -65,6 +70,7 @@ def mean_average_precision(pred_boxes, true_boxes, iou_threshold=0.5, box_format
         true_boxes (list): Similar as pred_boxes except all the correct ones
         iou_threshold (float): threshold where predicted bboxes is correct
         box_format (str): "midpoint" or "corners" used to specify bboxes
+        interpolate_pr: if to use interpolated pr curve, only effect plot not score
     Returns:
         float: mAP value across all classes given a specific IoU threshold
     """
@@ -73,6 +79,7 @@ def mean_average_precision(pred_boxes, true_boxes, iou_threshold=0.5, box_format
     label_wise_precision = nltk.defaultdict(float)
     label_wise_recall = nltk.defaultdict(float)
     label_wise_ap = nltk.defaultdict(float)
+    label_wise_pr_curve = dict()
 
     # list storing all AP for respective classes
     average_precisions = []
@@ -119,7 +126,7 @@ def mean_average_precision(pred_boxes, true_boxes, iou_threshold=0.5, box_format
             continue
         for detection_idx, detection in enumerate(detections):
             # Only take out the ground_truths that have the same
-            # training idx as detection
+            # training idx as detection (same image)
             ground_truth_img = [bbox for bbox in ground_truths if bbox[0] == detection[0]]
             num_gts = len(ground_truth_img)
             best_iou = 0
@@ -153,18 +160,31 @@ def mean_average_precision(pred_boxes, true_boxes, iou_threshold=0.5, box_format
         recalls = TP_cumsum / (total_true_bboxes + epsilon)
         precisions = TP_cumsum / (TP_cumsum + FP_cumsum + epsilon)
         precisions = torch.cat((torch.tensor([1]), precisions))
+
+        if interpolate_pr:
+            precisions = precisions.tolist()
+            for ipidx, _ in enumerate(precisions):
+                precisions[ipidx] = max(precisions[ipidx:])
+        precisions = torch.tensor(precisions)
+
         recalls = torch.cat((torch.tensor([0]), recalls))
+        label_wise_pr_curve[c] = {
+            'precisions': precisions.tolist() + [0],
+            'recalls': recalls.tolist() + [recalls.tolist()[-1]]
+        }
         # torch.trapz for numerical integration
         if precisions.shape[0] == 1:  # this means we never predict for this label
             label_wise_precision[c] = 0
         else:
             label_wise_precision[c] = precisions.numpy()[-1]
         label_wise_recall[c] = recalls.numpy()[-1]
+        # Note that trapz naturally calculated interpolated curve area
         label_ap = float(torch.trapz(precisions, recalls))
         # average_precisions.append(label_ap)
         label_wise_ap[c] = label_ap
 
-    return dict(label_wise_precision), dict(label_wise_recall), dict(label_wise_ap)
+    return dict(label_wise_precision), dict(label_wise_recall
+                                            ), dict(label_wise_ap), label_wise_pr_curve
 
 
 class label_metrics:
@@ -243,18 +263,103 @@ class label_metrics:
         self.pred_boxes = pred_boxes
         self.true_boxes = true_boxes
 
-    def get_metrics(self, iou_threshold=0.5):
-        label_wise_precision, label_wise_recall, label_wise_ap = mean_average_precision(
-            self.pred_boxes, self.true_boxes, iou_threshold=iou_threshold, box_format="midpoint"
-        )
-        return {
-            'precision': label_wise_precision,
-            'recall': label_wise_recall,
-            'average precision': label_wise_ap
-        }
+    def get_metrics(
+        self, iou_thresholds=[0.5, 0.7, 0.9], interpolate_pr=True, output_dir='eval_results'
+    ):
+        def uniquify(path):
+            filename, extension = os.path.splitext(path)
+            counter = 1
+
+            while os.path.exists(path):
+                path = filename + " (" + str(counter) + ")" + extension
+                counter += 1
+            return path
+
+        output_dir = uniquify(output_dir)
+        os.mkdir(output_dir)
+
+        table_result = []
+        column_names = []
+
+        label_all_precisions = nltk.defaultdict(list)
+        label_all_recalls = nltk.defaultdict(list)
+        label_all_f1 = nltk.defaultdict(list)
+        label_all_aps = nltk.defaultdict(list)
+        label_rows = nltk.defaultdict(list)
+
+        label_names = []
+
+        for iou_threshold in iou_thresholds:
+            print(f'processing IoU@{iou_threshold}')
+            column_names.extend([
+                f'P@{iou_threshold}', f'R@{iou_threshold}', f'F1@{iou_threshold}',
+                f'AP@{iou_threshold}'
+            ])
+            label_wise_precision, label_wise_recall, label_wise_ap, label_wise_pr_curve = mean_average_precision(
+                self.pred_boxes,
+                self.true_boxes,
+                iou_threshold=iou_threshold,
+                box_format="midpoint",
+                interpolate_pr=interpolate_pr
+            )
+
+            # Generate PR curve
+            plt.figure(figsize=(15, 15))
+            for label, prs in label_wise_pr_curve.items():
+                plt.xlim([0, 1])
+                plt.ylim([0, 1])
+                plt.plot(
+                    prs['recalls'],
+                    prs['precisions'],
+                    linewidth=2,
+                    label=f'{label} AP={label_wise_ap[label]:.2f}'
+                )
+                plt.xlabel('recall', fontsize=30)
+                plt.ylabel('precision', fontsize=30)
+                plt.gca().set_aspect('equal')
+            plt.plot([0, 1], [0, 1], '--', linewidth=3, color='red')
+            plt.legend(fontsize=25, loc='lower right')
+            plt.title(f'precision-recall IoU={iou_threshold}', fontsize=30)
+            plt.savefig(os.path.join(output_dir, f'precision-recall IoU={iou_threshold}.png'))
+            plt.close()
+
+            if len(label_names) > 0:
+                pass
+            else:
+                label_names = sorted(list(label_wise_precision.keys()))
+
+            for label in label_names:
+                label_p = label_wise_precision[label]
+                label_r = label_wise_recall[label]
+                label_f1 = 0.5 * label_p + 0.5 * label_r
+                label_ap = label_wise_ap[label]
+                label_all_precisions[label].append(label_p)
+                label_all_recalls[label].append(label_r)
+                label_all_f1[label].append(label_f1)
+                label_all_aps[label].append(label_ap)
+                label_rows[label].extend([label_p, label_r, label_f1, label_ap])
+
+        column_names.extend([f'P@All', f'R@All', f'F1@All', f'AP@All'])
+        for label in label_names:
+            row = label_rows[label] + [
+                sum(label_all_precisions[label]) / len(label_all_precisions[label]),
+                sum(label_all_recalls[label]) / len(label_all_recalls[label]),
+                sum(label_all_f1[label]) / len(label_all_f1[label]),
+                sum(label_all_aps[label]) / len(label_all_aps[label])
+            ]
+
+            table_result.append(row)
+
+        last_row = np.sum(table_result, axis=0) / len(table_result)
+        table_result.append(last_row)
+
+        df = pd.DataFrame(table_result, columns=column_names)
+        df.insert(0, "label", label_names + ['All'])
+        df.to_csv(os.path.join(output_dir, 'summary.csv'), index=False)
+        return df
 
     def find_confuse_labels(
-            self, iou_threshold=0.1, pred_boxes=None, true_boxes=None, box_format='midpoint'
+        self, iou_threshold=0.1, pred_boxes=None, true_boxes=None, box_format='midpoint'
     ):
         """
         This will give an idea of how model confused between each labels
@@ -347,5 +452,5 @@ def demo():
     pred_txt_dir = 'yolo_cls/runs/detect/exp/labels'
     truth_txt_dir = 'linmao_camera_data_real/labels/test'
     lm = label_metrics(pred_txt_dir, truth_txt_dir)
-    lm.get_metrics(iou_threshold=0.5)
+    lm.get_metrics(iou_thresholds=[0.5, 0.7, 0.9], interpolate_pr=True)
     lm.find_confuse_labels(iou_threshold=0.1, box_format='midpoint')
